﻿using System;
using System.Diagnostics;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Description;
using Castle.DynamicProxy;
using System.ServiceModel.Channels;

namespace CastleAbstraction.Proxies
{

    /// <summary>
    /// Interceptor class
    /// </summary>
    [Serializable]
    public class ProxyInterceptor : IInterceptor
    {

        /// <summary>
        /// Intercepts the specified invocation.
        /// </summary>
        /// <param name="invocation">The invocation.</param>
        public void Intercept(IInvocation invocation)
        {

            // get constructor for ChannelFactory<T>
            var createFactoryMethod = typeof(ChannelFactory<>)
                .MakeGenericType(new[] { invocation.Method.ReflectedType })
                .GetConstructor(new[] { typeof(Binding), typeof(string) });
            Debug.Assert(createFactoryMethod != null, "createFactoryMethod != null");
            // Create instance of ChannelFactory
            using (var factory = (ChannelFactory)createFactoryMethod.Invoke(
                new object[]
                        {
                            new WebHttpBinding(), 
                            ProxyFactory.Url
                        }))
            {
                factory.Endpoint.Behaviors.Add(new WebHttpBehavior());
                // Call remote service
                invocation.ReturnValue = Execute(invocation, factory);
            }

        }

        /// <summary>
        /// Executes the specified method on a service.
        /// </summary>
        /// <param name="invocation">The invocation.</param>
        /// <param name="factory">The factory.</param>
        /// <returns></returns>
        private object Execute(IInvocation invocation, ChannelFactory factory)
        {
            try
            {
                // Get method to create channel
                var channel = factory.GetType()
                                     .GetMethod("CreateChannel", new Type[] { })
                                     .Invoke(factory, new object[] { });
                Debug.Assert(channel != null, "channel != null");
                object returnValue;
                //Create scope - this is optional and needed if you use cookies on services
                using (var scope = new OperationContextScope((IContextChannel)channel))
                {
                    // may need to add null checking
                    // FInd remote service method to invoke
                    var methodToInvoke = channel.GetType().GetMethod(
                        invocation.Method.Name,
                        invocation.Arguments.Select(argument => argument.GetType()).ToArray());
                    // call the method
                    returnValue = methodToInvoke.Invoke(channel, invocation.Arguments);
                    // close factory
                    factory.Close();
                }

                return returnValue;
            }
            catch
            {
                // abort factory in case of an error to release resources.
                if (factory != null) factory.Abort();
                throw;
            }
        }
    }

}